import sys
import time
import jsonschema

from pyneval.metric.utils.config_utils import DINF
from pyneval.model.swc_node import SwcTree
from pyneval.metric.utils.km_utils import KM, get_dis_graph
from pyneval.io.read_json import read_json
from pyneval.metric.utils.point_match_utils import get_swc2swc_dicts


def get_neighbors(node):
    neighbors = list(node.children)
    if node.parent is not None and not node.parent.is_virtual():
        neighbors.append(node.parent)
    return neighbors


def get_same_pos_nodes(gold_nodes, test_nodes):
    res = []
    for gold_node in gold_nodes:
        for test_node in test_nodes:
            if gold_node.get_center_as_tuple() == test_node.get_center_as_tuple():
                res.append(gold_node)
                break
    return res


def get_extra_pos_nodes(big_set, subset):
    res = []
    pos_set = set()
    for node in subset:
        pos_set.add(node.get_center_as_tuple())

    for node in big_set:
        if not node.get_center_as_tuple() in pos_set:
            res.append(node)
    return res


def link_metric(gold_swc_tree, test_swc_tree, config):
    scale = config['scale']
    gold_swc_tree.rescale(scale)
    test_swc_tree.rescale(scale)

    gold_list = gold_swc_tree.get_node_list()
    test_list = test_swc_tree.get_node_list()

    gold_test_dict = get_swc2swc_dicts(src_node_list=gold_swc_tree.get_node_list(),
                                       tar_node_list=test_swc_tree.get_node_list())
    test_gold_dict = get_swc2swc_dicts(src_node_list=test_swc_tree.get_node_list(),
                                       tar_node_list=gold_swc_tree.get_node_list())

    edge_loss, tree_dis_loss = 0.0, 0.0
    for gold_node in gold_list:
        if gold_node.is_virtual():
            continue

        test_node = gold_test_dict[gold_node]

        # attention: nodes in test neighbors are labeled by TEST tree label
        test_neighbors = get_neighbors(test_node)
        gold_neighbors = get_neighbors(gold_node)

        nodes_in_same_pos = get_same_pos_nodes(gold_neighbors, test_neighbors)
        nodes_gold_extra = get_extra_pos_nodes(gold_neighbors, nodes_in_same_pos)
        # attention: nodes in test neighbors are labeled by TEST tree label
        nodes_test_extra = get_extra_pos_nodes(test_neighbors, nodes_in_same_pos)

        for gold_extra in nodes_gold_extra:
            edge_loss += gold_extra.distance(gold_node)

        for test_extra in nodes_test_extra:
            map_test_extra = test_gold_dict[test_extra]
            edge_loss += map_test_extra.distance(gold_node)

        # if two nodes are not connected, the dis will be set as DINF, so the threshold need to be set as DINF - 1
        dis_graph, switch, test_len, gold_len = \
            get_dis_graph(gold_tree=gold_swc_tree, test_tree=test_swc_tree,
                          gold_node_list=nodes_gold_extra, test_node_list=nodes_test_extra,
                          test_gold_dict=test_gold_dict, threshold_dis=DINF-1, metric_mode=2)

        km = KM(maxn=max(test_len, gold_len) + 10, nx=test_len, ny=gold_len, G=dis_graph)
        km.solve()
        tree_dis_loss += -km.get_max_dis()

    res = {
        "edge_loss": edge_loss,
        "tree_dis_loss": tree_dis_loss
    }
    return res, None, None


if __name__ == "__main__":
    sys.setrecursionlimit(1000000)
    start = time.time()
    gold_swc_tree = SwcTree()
    test_swc_tree = SwcTree()
    test_swc_tree.load("..\\..\\data\\test_data\\topo_metric_data\\gold_fake_data4.swc")
    gold_swc_tree.load("..\\..\\data\\test_data\\topo_metric_data\\test_fake_data4.swc")
    config = read_json("..\\..\\config\\link_metric.json")
    config_schema = read_json("..\\..\\config\\schemas\\link_metric_schema.json")

    try:
        jsonschema.validate(config, config_schema)
    except Exception as e:
        raise Exception("[Error: ]Error in analyzing config json file")

    link_res = link_metric(test_swc_tree=test_swc_tree, gold_swc_tree=gold_swc_tree, config=config)
    print("edge loss     = {}\n"
          "tree_dis_loss = {}".format(link_res["edge_loss"], link_res["tree_dis_loss"]))



import os

from pyneval.tools.optimize.SA import SAFast
from pyneval.metric import ssd_metric
from pyneval.model import swc_node
from pyneval.io import read_json
from scipy import stats as st

import copy
import time
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

g_score = -1
g_gold_tree = None
g_tar_tree = None
g_rcn_method = None
g_rcn_config = None
g_metric_method = None
g_metric_configs = None

# Gold file and Test tif need to placed in ../../../data/optimation
# and named after FILE_ID_test.tif, FILE_ID_gold.swc
FILE_ID = "test1"
NEUTU_PATH = "neutu"
ORIGIN_PATH = "../../../data/optimation/{}/{}_test.tif".format(FILE_ID, FILE_ID)
GOLD_PATH = "../../../data/optimation/{}/{}_gold.swc".format(FILE_ID, FILE_ID)
METRIC_CONFIG_PATH = "../../../config/ssd_metric.json"
LOG_PATH = "../../../output/optimization/neutu_log.txt"
# specific test name is given in SA.py
TEST_PATH = "../../../data/optimation/output/"
CONFIG_PATH = "../../../config/fake_reconstruction_configs/"



def SA_optimize(configs=None, test_name=None, lock=None):
    global g_metric_method
    global g_metric_configs
    global g_rcn_config
    # identify specific TEST output path and CONFIG input PATH
    LOC_TEST_PATH = os.path.join(TEST_PATH, test_name+"_test.swc")
    rec_config = copy.deepcopy(g_rcn_config)

    if configs is not None:
        LOC_CONFIG_PATH = os.path.join(CONFIG_PATH, test_name+".json")
        rec_config["trace"]["default"]["minimalScoreAuto"] = configs[0]
        rec_config["trace"]["default"]["minimalScoreSeed"] = configs[1]

        read_json.save_json(LOC_CONFIG_PATH, rec_config)
    else:
        LOC_CONFIG_PATH = os.path.join(CONFIG_PATH, "best_x_{}.json".format(test_name))

    REC_CMD = "{} --command --trace {} -o {} --config {} > {}".format(
        NEUTU_PATH, ORIGIN_PATH, LOC_TEST_PATH, LOC_CONFIG_PATH, LOG_PATH
    )
    try:
        os.system(REC_CMD)
    except:
        raise Exception("[Error: ] error executing reconstruction")

    res_tree = swc_node.SwcTree()
    gold_tree = swc_node.SwcTree()
    res_tree.load(LOC_TEST_PATH)
    gold_tree.load(GOLD_PATH)

    if lock is not None:
        lock.acquire()
    try:
        main_score = g_metric_method(gold_tree, res_tree, g_metric_configs)
    finally:
        if lock is not None:
            lock.release()
    score = (main_score["recall"] + main_score["precision"])/2
    print("[Info: ] ssd loss = {}".format(score))

    return configs, -score


def main():
    global g_metric_configs
    global g_metric_method
    global g_rcn_config
    g_metric_method = ssd_metric.ssd_metric
    g_metric_configs = read_json.read_json(METRIC_CONFIG_PATH)
    g_rcn_config = read_json.read_json(os.path.join(CONFIG_PATH, "default.json"))

    # optimize with SA
    # configs here is the config of the reconstruction
    configs = (0.3, 0.35)
    start = time.time()
    sa_fast = SAFast(func=SA_optimize,
                     x0=configs, T_max=0.01, T_min=1e-5, q=0.96, L=25, max_stay_counter=15, upper=1, lower=0)
    best_configs, best_value = sa_fast.run()
    print("[Info: ]best configs:\n"
          "        origin minimalScoreAuto = {}\n"
          "        minimalScoreSeed = {}\n"
          "        best value = {}\n"
          "        time = {}\n" .format(
        best_configs[0], best_configs[1], best_value, time.time() - start
    ))
    # save best json file
    g_rcn_config["trace"]["default"]["minimalScoreAuto"] = best_configs[0]
    g_rcn_config["trace"]["default"]["minimalScoreSeed"] = best_configs[1]
    read_json.save_json(os.path.join(CONFIG_PATH, "best_x_{}.json".format(FILE_ID)), g_rcn_config)
    # get and save best reconstruct swc
    print("[Info: ]Exam test score with best configs: ")
    cfg, score = SA_optimize(test_name = FILE_ID)
    # plot the result.
    plt.plot(pd.DataFrame(sa_fast.best_y_history).cummin(axis=0))
    plt.xlabel("iterations")
    plt.ylabel("score(opposite value)")
    plt.show()
    return 0


if __name__ == "__main__":
    main()

    # g_metric_method = ssd_metric.ssd_metric
    # g_metric_configs = read_json.read_json(METRIC_CONFIG_PATH)
    # g_rcn_config = read_json.read_json(os.path.join(CONFIG_PATH, "6656_2304_22016.json"))
    
    # SA_optimize(test_name="6656_2816_22016")



import argparse
import sys
import os
import jsonschema
import pyneval
from pyneval.io.read_swc import read_swc_trees
from pyneval.io import read_json
from pyneval.io.swc_writer import swc_save
from pyneval.io.read_tiff import read_tiffs
from pyneval.metric import diadem_metric
from pyneval.metric import length_metric
from pyneval.metric import volume_metric
from pyneval.metric import branch_leaf_metric
from pyneval.metric import link_metric
from pyneval.metric import ssd_metric
from pyneval.metric.utils import anno_utils


METRICS = {
    'diadem_metric': {
        'config': "diadem_metric.json",
        'description': "DIADEM metric (https://doi.org/10.1007/s12021-011-9117-y)",
        'alias': ['DM'],
        'method': diadem_metric.diadem_metric,
        'public': True
    },
    'ssd_metric': {
        'config': "ssd_metric.json",
        'description': "minimum square error between up-sampled gold and test trees",
        'alias': ['SM'],
        'method': ssd_metric.ssd_metric,
        'public': True
    },
    'length_metric': {
        'config': "length_metric.json",
        'description': "length of matched branches and fibers",
        'alias': ['ML'],
        'method': length_metric.length_metric,
        'public': True
    },
    'volume_metric': {
        'config': "volume_metric.json",
        'description': "volume overlap",
        'alias': ['VM'],
        'method': volume_metric.volume_metric,
        'public': False
    },
    'branch_metric': {
        'config': "branch_metric.json",
        'description': "quality of critical points",
        'alias': ['BM'],
        'method': branch_leaf_metric.branch_leaf_metric,
        'public': True
    },
    'link_metric': {
        'config': "link_metric.json",
        'description': "",
        'alias': ['LM'],
        'method': link_metric.link_metric,
        'public': False
    },
}

METRIC_ALIAS_MAP = {}

for metric in METRICS:
    if 'alias' in METRICS[metric]:
        for alias in METRICS[metric]['alias']:
            METRIC_ALIAS_MAP[alias] = metric

def get_root_metric(metric):
    if metric in METRIC_ALIAS_MAP:
        return METRIC_ALIAS_MAP[metric]
    elif metric in METRICS:
        return metric

def get_metric_config(metric):
    return METRICS[get_root_metric(metric)]

def get_metric_summary(with_description):
    summary = ''
    if with_description:
        for metric in METRICS:
            if METRICS[metric].get('public', False):
                description = METRICS[metric]['description']
                summary += '[{}]: '.format(metric) + (description if description else '[No description]') + '\n'
    else:
        summary = ', '.join((filter(lambda m: METRICS[m].get('public', False), METRICS.keys())))

    return summary

# For local setup: clone code and pip setup.py install
# config_dir = os.path.join(os.path.dirname(pyneval.__file__), '../config/pyneval', )
# For Pypi:
config_dir = os.path.join(sys.prefix, 'config/pyneval', )

def get_metric_config_path(metric, root_dir):
    return os.path.join(config_dir, get_metric_config(metric)['config'])

def get_metric_config_schema_path(metric, root_dir):
    schema_dir = os.path.join(config_dir, "schemas")
    return os.path.join(schema_dir, get_metric_config(metric)['config'][:-5]+"_schema.json")

def get_metric_method(metric):
    return get_metric_config(metric)['method']

def read_parameters():
    parser = argparse.ArgumentParser(
        description="pyneval 1.0"
    )

    parser.add_argument(
        "--gold",
        "-G",
        help="path of the gold standard SWC file",
        required=False
    )
    parser.add_argument(
        "--test",
        "-T",
        help="a list of reconstructed SWC files or folders for evaluation",
        required=False,
        nargs='*',
    )
    parser.add_argument(
        "--metric",
        "-M",
        help="metric choice: " + get_metric_summary(False) + ".",
        required=False
    )
    parser.add_argument(
        "--output",
        "-O",
        help="output path of metric results, output file is in json format with different scores of the metric",
        required=False
    )
    parser.add_argument(
        "--detail",
        "-D",
        help="output path of detail metric result, swc format presented.\n"
             "identify different type according to metric result for each node",
        required=False
    )
    parser.add_argument(
        "--config",
        "-C",
        help="path of custom configuration file for the specified metric",
        required=False
    )
    parser.add_argument(
        "--debug",
        help="print debug info or not",
        required=False
    )
    return parser.parse_args()


def init(abs_dir):
    sys.path.append(abs_dir)
    sys.path.append(os.path.join(abs_dir, "src"))
    sys.path.append(os.path.join(abs_dir, "test"))
    sys.setrecursionlimit(1000000)


def set_configs(abs_dir, args):
    # argument: gold
    gold_swc_path = os.path.join(abs_dir, args.gold)
    if not (os.path.isfile(gold_swc_path) and gold_swc_path[-4:] == ".swc"):
        raise Exception("[Error: ] gold standard file is not a swc file")
    gold_swc_tree = read_swc_trees(gold_swc_path)[0]  # SwcTree

    # argument: metric
    metric = get_root_metric(args.metric)
    if not metric:
        raise Exception("\nERROR: The metric '{}' is not supported.".format(args.metric) +
                        "\nValid options for --metric:\n" +
                        get_metric_summary(True))

    # argument: test
    test_swc_paths = [os.path.join(abs_dir, path) for path in args.test]
    test_swc_trees = []
    # read test trees
    if metric in ['volume_metric', 'VM']:
        for file in test_swc_paths:
            test_swc_trees += read_tiffs(file)
    else:
        for file in test_swc_paths:
            test_swc_trees += read_swc_trees(file)

    # info: how many trees read
    print("There are {} test image(s)".format(len(test_swc_trees)))

    # argument: config
    config_path = args.config
    if config_path is None:
        config_path = get_metric_config_path(metric, abs_dir)
    config_schema_path = get_metric_config_schema_path(metric, abs_dir)

    config = read_json.read_json(config_path)
    config_schema = read_json.read_json(config_schema_path)

    jsonschema.validate(config, config_schema)


    # argument: output
    output_dir = None
    if args.output:
        output_dir = os.path.join(abs_dir, args.output)

    # argument: detail
    detail_dir = None
    if args.detail:
        detail_dir = os.path.join(abs_dir, args.detail)
        config["detail"] = True

    # argument: debug
    is_debug = args.debug

    return gold_swc_tree, test_swc_trees, metric, output_dir, detail_dir, config, is_debug


def excute_metric(metric, gold_swc_tree, test_swc_tree, config, detail_dir, output_dir):
    metric_method = get_metric_method(metric)
    test_swc_name = test_swc_tree.get_name()

    result, res_gold_swc_tree, res_test_swc_tree = metric_method(gold_swc_tree=gold_swc_tree,
                                                                 test_swc_tree=test_swc_tree, config=config)

    print("---------------Result---------------")
    for key in result:
        print("{} = {}".format(key.ljust(15, ' '), result[key]))
    print("----------------End-----------------\n")

    file_name = test_swc_name[:-4] + "_" + metric + "_"

    if detail_dir:
        if res_gold_swc_tree is not None:
            swc_save(swc_tree=res_gold_swc_tree,
                     out_path=os.path.join(detail_dir, file_name + "recall.swc"),
                     extra=anno_utils.get_detail_type(metric))
        if res_test_swc_tree is not None:
            swc_save(swc_tree=res_test_swc_tree,
                     out_path=os.path.join(detail_dir, file_name + "precision.swc"),
                     extra=anno_utils.get_detail_type(metric))

    if output_dir:
        read_json.save_json(data=result,
                            json_file_path=os.path.join(output_dir, file_name + ".json"))


# command program
def run():
    abs_dir = os.path.abspath("")
    init(abs_dir)

    args = read_parameters()
    gold_swc_tree, test_swc_trees, metric, output_dir, detail_dir, config, is_debug = set_configs(abs_dir, args)

    for test_swc_tree in test_swc_trees:
        excute_metric(metric=metric, gold_swc_tree=gold_swc_tree, test_swc_tree=test_swc_tree,
                      config=config, detail_dir=detail_dir, output_dir=output_dir)


if __name__ == "__main__":
    sys.exit(run())
    # print(pyneval.__file__)

# pyneval --test D:\gitProject\mine\PyNeval\test\data_example\test\2_18_test.swc --gold D:\gitProject\mine\PyNeval\test\data_example\gold\2_18_gold.swc --metric matched_length --reverse true

# pyneval --test D:\gitProject\mine\PyNeval\test\data_example\test\194444.swc --gold D:\gitProject\mine\PyNeval\test\data_example\gold\194444.swc --metric matched_length --reverse true

# pyneval --test D:\gitProject\mine\PyNeval\test\data_example\test\diadem\diadem1.swc --gold D:\gitProject\mine\PyNeval\test\data_example\gold\diadem\diadem1.swc --metric diadem_metric

# pyneval --test D:\gitProject\mine\PyNeval\test\data_example\test\diadem\diadem7.swc --gold D:\gitProject\mine\PyNeval\test\data_example\gold\diadem\diadem7.swc --metric diadem_metric

# pyneval --gold .\\data\example_selected\g.swc --test .\\data\example_selected\g.tif --metric volume_metric --output D:\gitProject\mine\PyNeval\output\volume_metric\volume_out.swc

# pyneval --gold .\\data\test_data\topo_metric_data\gold_fake_data1.swc --test .\data\test_data\topo_metric_data\test_fake_data1.swc --metric link_metric

# pyneval --gold .\\data\\test_data\\ssd_data\\gold\\a.swc --test .\\data\\test_data\\ssd_data\\test\\a.swc --metric branch_metric

# pyneval --gold .\\data\test_data\geo_metric_data\gold_34_23_10.swc --test .\data\test_data\geo_metric_data\test_34_23_10.swc --metric branch_metric

# pyneval --gold ./data/test_data/geo_metric_data/gold_fake_data1.swc --test ./data/test_data/geo_test/test_fake_data1.swc --metric branch_metric --detail ./output/detail --output ./output/output



import os
import platform


def read_int_config(config, config_name, default):
    config_value = 0.0
    if config_name not in config.keys() or config[config_name] == "default":
        print("[Warning: ]config {} is not defined or value is default, default value {} is used".format(
            config_name, default))
        config_value = default
    else:
        config_value = int(config[config_name])
    return config_value


def read_float_config(config, config_name, default):
    config_value = 0.0
    if config_name not in config.keys() or config[config_name] == "default":
        # print("[Warning: ]config {} is not defined or value is default, default value {} is used".format(
        #     config_name, default))
        config_value = default
    else:
        config_value = float(config[config_name])
    return config_value


def read_path_config(config, config_name, default):
    config_value = ""
    if config_name in config:
        config_value = config[config_name]
        if platform.system() == "Linux":
            config_value = '/'.join(config_value.split("\\"))
    else:
        # print("[Warning: ]config {} is not defined, default value {} is used".format(config_name, default))
        config_value = default
    return config_value


def read_bool_config(config, config_name, default):
    config_value = 0.0
    if config_name not in config.keys() or config[config_name] == "default":
        # print("[Warning: ]config {} is not defined or value is default, default value {} is used".format(
        #     config_name, default))
        config_value = default
    else:
        config_value = bool(config[config_name])
    return config_value


from anytree import NodeMixin
from pyneval.model.euclidean_point import EuclideanPoint
import copy
import math
import queue

_3D = "3d"
_2D = "2d"
TRAJECTORY_NONE = -1.0
FLOAT_INF = 999999999.9
DEBUG = False
RIGHT = 'right'
LEFT = 'left'
DEFULT = 'DEFULT'


class BinaryNode(NodeMixin):
    """
        this is a class that temporarily store binarytree transformed from swcTree
        Attributes:
        id: id of the node,
        type: leaf = 1,continuation = 2, bifurcation = 3,
        parent: pa node id,
        son=[]: son list,
        x: x coordinate,
        y: y coordinate,
        z: z coordinate,
        radius: radius of the node
    """

    def __init__(self,
                 data=object,
                 parent=None,
                 left_son=None,
                 right_son=None,

                 max_dep=0,
                 hight=0,
                 treesize=1,
                 leaves=1):
        self.data=data
        self.parent=parent
        self.left_son=left_son
        self.right_son=right_son
        self.max_dep=max_dep
        self.hight=hight
        self.treesize=treesize

    def has_children(self):
        if self.left_son is not None or self.right_son is not None:
            return True
        return False

    def is_left(self):
        if self.get_side() == LEFT:
            return True
        return False

    def get_side(self):
        if self.is_root():
            return DEFULT
        if self.parent.left_son == self:
            return LEFT
        if self.parent.right_son == self:
            return RIGHT
        return LEFT

    def is_root(self):
        if self.parent == None:
            return True
        return False

    def is_leaf(self):
        if self.left_son == None and self.right_son == None:
            return True
        return False

    def to_str(self):
        print("id = {} path_length= {} xy_path_length = {} z_path_length = {}".format(
            self.data.get_id(), self.data.path_length, self.data.xy_path_length, self.data.z_path_length))
        if self.left_son is not None:
            print("left son: {}".format(self.left_son.data.get_id()))
        if self.right_son is not None:
            print("right son {}".format(self.right_son.data.get_id()))
        print("---------------------------")

    def print_tree(self):
        q = queue.LifoQueue()
        q.put(self)
        while not q.empty():
            cur = q.get()
            if cur is None:
                continue
            cur.to_str()
            q.put(cur.left_son)
            q.put(cur.right_son)

    def get_node_list(self):
        node_list = list()
        stack = queue.LifoQueue()
        stack.put(self)

        while not stack.empty():
            node = stack.get()
            node_list.append(node)
            if node.has_children():
                stack.put(node.left_son)
                stack.put(node.right_son)
        return node_list



import json
import os,sys


def read_json(json_file_path, DEBUG=False):
    json_file_path = os.path.normpath(json_file_path)
    if not os.path.isfile(json_file_path) or not (json_file_path[-5:] == ".json" or json_file_path[-5:] == ".JSON"):
        try:
            data = json.loads(json_file_path)
            if DEBUG:
                print(data)
            return data
        except:
            raise Exception("[Error: ] \" {} \"  is not a json file. Wrong format".format(json_file_path))

    with open(json_file_path,'r') as f:
        data = json.load(f)
        if DEBUG:
            print(type(data))
        return data


def save_json(json_file_path, data, DEBUG=False):
    json_file_path = os.path.normpath(json_file_path)
    if not(json_file_path[-5:] == ".json" or json_file_path[-5:] == ".JSON"):
        raise Exception("[Error: ] \" {} \"  is not a json file. Wrong format".format(json_file_path))
    try:
        with open(json_file_path, 'w') as f:
            json.dump(data, f, indent=4)
            if DEBUG:
                print(type(data))
    except:
        raise Exception("[Error: ] json file save error")
    return True


if __name__ == "__main__":
    # read_json(r'{"method": 2,"thereshold": "default"')
    test_json = read_json("/home/zhanghan/01_project/Pyneval/config/schemas/branch_metric_schema.json")
    save_json("./test.json", test_json)



import queue
import time
import math
import numpy as np
import jsonschema

from anytree import PreOrderIter
from pyneval.io import read_json
from pyneval.io import read_swc
from pyneval.model import binary_node
from pyneval.model import swc_node
from pyneval.metric.utils import config_utils
from pyneval.metric.utils import diadam_match_utils
from pyneval.metric.utils import bin_utils
from pyneval.metric.utils import point_match_utils


# thresholds
g_terminal_threshold = 0

WEIGHT_DEGREE = 1
WEIGHT_SQRT_DEGREE = 2
WEIGHT_DEGREE_HARMONIC_MEAN = 3
WEIGHT_PATH_LENGTH = 4
WEIGHT_UNIFORM = 5
_2D = '2d'
_3D = '3d'

# switch
g_remove_spur = False
g_align_tree_by_root = False
g_count_excess_nodes = True
g_weight_node = WEIGHT_DEGREE
g_list_miss = False
g_list_distant_matches = False
g_list_continuations = False
g_find_proper_root = False

# various
g_weight_sum = 0
g_score_sum = 0
g_quantity_score_sum = 0
g_quantity_score = 0
g_direct_match_score = 0
g_final_score = -1

g_miss = set()
g_spur_set = set()
g_matches = {}
g_weight_dict = {}
g_excess_nodes = {}
g_distance_match = []
g_continuation = []


def remove_spurs(bin_root, threshold):
    spur_set = set()
    both_children_spur = False

    stack = queue.LifoQueue()

    if bin_root.has_children:
        stack.put(bin_root)

    while not stack.empty():
        node = stack.get()
        if node.has_children():
            stack.put(node.left_son)
            stack.put(node.right_son)
        else:
            distance = node.data.path_length
            if distance < threshold and not (node in g_matches):
                both_children_spur = False
                spur_set.add(node)
                if node.get_side() == binary_node.RIGHT:
                    l_node = node.parent.left_son
                    if not l_node.has_children and \
                            l_node.data.path_length < threshold and \
                            not l_node in g_matches:
                        both_children_spur = True
                        spur_set.add(l_node)
                        ll_node = stack.get()
                        if ll_node != l_node:
                            raise Exception("[Error:  ] stack top is not current node's twin")

                if not both_children_spur:
                    spur_set.add(node.parent)
    return spur_set


def generate_node_weights(bin_root, bin_subroot, spur_set, DEBUG=False):
    init_stack = queue.LifoQueue()
    main_stack = queue.LifoQueue()
    degree_dict = {}
    global g_weight_dict
    degree = 0
    for subroot in bin_subroot:
        init_stack.put(subroot)
        main_stack.put(subroot)

        while not init_stack.empty():
            node = init_stack.get()
            if node.has_children():
                init_stack.put(node.left_son)
                init_stack.put(node.right_son)
                main_stack.put(node.left_son)
                main_stack.put(node.right_son)

        while not main_stack.empty():
            node = main_stack.get()

            if g_weight_node == WEIGHT_UNIFORM:
                g_weight_dict[node] = 1
            else:
                if DEBUG:
                    print("Determining weight for {}".format(node.data.id))

                if node.is_leaf():
                    g_weight_dict[node] = 1
                    degree_dict[node] = 1
                else:
                    degree = 0
                    if node.left_son.is_leaf() and node.right_son.is_leaf():
                        if node.left_son in spur_set or node.right_son in spur_set:
                            degree = 1
                        else:
                            degree = 2
                    elif node in spur_set:
                        if node.left_son.is_leaf():
                            degree += degree_dict[node.right_son]
                        else:
                            degree += degree_dict[node.left_son]
                    else:
                        degree += degree_dict[node.left_son]
                        degree += degree_dict[node.right_son]
                    degree_dict[node] = degree
                    if g_weight_node == WEIGHT_DEGREE:
                        g_weight_dict[node] = degree
                    elif g_weight_node == WEIGHT_SQRT_DEGREE:
                        g_weight_dict[node] = math.sqrt(degree)
                    elif g_weight_node == WEIGHT_DEGREE_HARMONIC_MEAN:
                        g_weight_dict[node] = 2.0 / (1.0/degree_dict[node.getLeft()] + 1.0/degree_dict.get[node.getRight()])
                    elif g_weight_node == WEIGHT_PATH_LENGTH:
                        g_weight_dict[node] = node.data.path_length
                    if DEBUG:
                        print("nodeId = {}, degree = {}".format(node.data._id, degree))
                        if node.parent is not None:
                            print("pa = {}".format(node.parent.data._id))
                        print("-----------")


def is_diadem_match(gold_node, nearby_node, bin_gold_list, bin_test_list, DEBUG=False):
    length_diff = diadam_match_utils.get_match_path_length_difference(gold_node,
                                                                      nearby_node,
                                                                      bin_gold_list,
                                                                      bin_test_list)
    if DEBUG:
        print("gold_node = {}, test_node = {}, length_diff = {}".format(
            gold_node.data.get_id(), nearby_node.data.get_id(), length_diff
        ))
    if length_diff < 1:
        return True
    else:
        return False


def get_best_match(match_list, gold_node, bin_test_list,DEBUG=False):
    if DEBUG:
        print("Determining best match")
    confirm_list = []

    gold_data = gold_node.data
    des_branch_left = {}
    trajactory_map = {}

    if gold_node.has_children():
        nearby_list = []
        path_length_map = {}
        path_length_map[gold_node] = swc_node.SwcNode()

        target_dis_list = []
        for test_node in match_list:
            target_distances = swc_node.SwcNode()
            target_distances.xy_path_length = gold_node.data.distance(test_node.data, _2D)
            target_distances.z_path_length = gold_node.data.distance(test_node.data, _2D)
            target_dis_list.append(target_distances)

        current_list = []
        current_list.append(gold_node.left_son)
        trajactory_map[gold_node.left_son] = gold_node.data.left_trajectory
        des_branch_left[gold_node.left_son] = True

        current_list.append(gold_node.right_son)
        trajactory_map[gold_node.right_son] = gold_node.data.right_trajectory
        des_branch_left[gold_node.right_son] = False

        while len(current_list) > 0 and len(confirm_list) == 0:
            next_list = []
            for des_node in current_list:
                target_tra = trajactory_map[des_node]

                if target_tra.get_x() == -1.0 or target_tra.get_z() == -1.0:
                    tmp_tra = diadam_match_utils.get_trajectory_for_path(gold_node, des_node)
                    if target_tra.get_x() == -1.0:
                        target_tra.set_x(tmp_tra.get_x())
                        target_tra.set_y(tmp_tra.get_y())
                    if target_tra.get_z() == -1.0:
                        target_tra.set_z(tmp_tra.get_z())
                des_tra = des_node.data.parent_trajectory

                if des_node.has_children():
                    next_list.append(des_node.left_son)
                    next_list.append(des_node.right_son)
                    trajactory_map[des_node.left_son] = target_tra
                    trajactory_map[des_node.right_son] = target_tra

                gold_path_length = swc_node.SwcNode()
                gold_path_length.add_length(des_node.data)
                parent_path_length = path_length_map[des_node.parent]
                gold_path_length.add_length(parent_path_length)
                path_length_map[des_node] = gold_path_length

                nearby_list = diadam_match_utils.get_nearby_node_list(des_node,
                                                                      bin_test_list,
                                                                      check_previous_use=True,
                                                                      g_matches=g_matches)

                for test_node in nearby_list:
                    test_des_node = test_node
                    test_path_length = swc_node.SwcNode()

                    des_dis = swc_node.SwcNode()
                    des_dis.xy_path_length = des_node.data.distance(test_node.data, _2D)
                    des_dis.z_path_length = math.fabs(des_node.data.get_z() - test_node.data.get_z())

                    while test_node.parent is not None:
                        test_path_length.add_length(test_node.data)
                        test_node = test_node.parent
                        tmp_list = []
                        for match in match_list:
                            if test_node == match:
                                test_XY_path_length = test_path_length.xy_path_length + \
                                                      diadam_match_utils.get_end_node_XY_dis_diff(gold_data, target_tra, test_node.data) + \
                                                      diadam_match_utils.get_end_node_XY_dis_diff(des_node.data, des_tra, test_des_node.data)
                                test_Z_path_length = test_path_length.z_path_length + \
                                                     diadam_match_utils.get_end_node_Z_dis_diff(gold_data, target_tra, test_node.data) + \
                                                     diadam_match_utils.get_end_node_Z_dis_diff(des_node.data, des_tra, test_des_node.data)

                                if diadam_match_utils.path_length_matches(gold_path_length,
                                                                          test_XY_path_length,
                                                                          test_Z_path_length) < 1:
                                    confirm_list.append(match)
                                    tmp_list.append(match)

                        for node in tmp_list:
                            match_list.remove(node)
                        while len(tmp_list):
                            tmp_list.pop()

            current_list = nearby_list
    if len(confirm_list) == 1:
        return confirm_list[0]

    if len(confirm_list) == 0:
        confirm_list = match_list

    dis = 0
    closest_dis = -1
    closest_match = None
    for test_node in confirm_list:
        dis = test_node.data.distance(gold_node.data)
        if closest_dis == -1 or dis < closest_dis:
            closest_dis = dis
            closest_match = test_node

    return closest_match


def get_closest_match(gold_node,
                      bin_gold_list,
                      bin_test_list,
                      DEBUG=False):
    best_match = None

    # Step1: find node within threshold
    nearby_list = diadam_match_utils.get_nearby_node_list(gold_node=gold_node,
                                                          bin_test_list=bin_test_list,
                                                          check_previous_use=True,
                                                          g_matches=g_matches)
    if DEBUG:
        print("nearby list of {}:".format(gold_node.data.get_id()))
        for node in nearby_list:
            print(node.data.get_id())
        print("----END----")

    # Step2: find match node whose ancestor is also matched
    match_list = []
    for nearby_node in nearby_list:
        if is_diadem_match(gold_node, nearby_node, bin_gold_list, bin_test_list):
            if DEBUG:
                print("{} {}".format(
                    gold_node.data.get_id(), nearby_node.data.get_id()
                ))
            match_list.append(nearby_node)

    if len(match_list) == 1:
        best_match = match_list[0]
    elif len(match_list) >= 1:
        best_match = get_best_match(match_list, gold_node, bin_test_list)

    if DEBUG:
        if best_match is not None:
            print("best match {} {}".format(gold_node.data.get_id(), best_match.data.get_id()))
        else:
            print("best match {} None".format(gold_node.data.get_id()))
    return best_match


def get_des_in_for_continuation(first_node, ancestor_node, ancestor_match,
                                ancestor_trajectory,path_length_map,bin_test_list,DEBUG=False):
    spe_ancestor_trajectory = ancestor_trajectory
    test_matches = []

    stack = queue.LifoQueue()
    stack.put(first_node)
    while not stack.empty():
        gold_node = stack.get()
        des_tra = gold_node.data.parent_trajectory
        # print("gold_node id = {}".format(gold_node.data.get_id()))
        if gold_node in g_matches:
            test_matches.append(g_matches[gold_node])
        else:
            test_matches = diadam_match_utils.get_nearby_node_list(gold_node,
                                                                   bin_test_list,
                                                                   check_previous_use=False)

        prev_path_length = path_length_map[gold_node.parent]
        gold_path_length = swc_node.SwcNode()
        gold_path_length.path_length = gold_node.data.path_length + prev_path_length.path_length
        gold_path_length.xy_path_length = gold_node.data.xy_path_length + prev_path_length.xy_path_length
        gold_path_length.z_path_length = gold_node.data.z_path_length + prev_path_length.z_path_length
        path_length_map[gold_node] = gold_path_length

        for child_match in test_matches:
            test_path_length = swc_node.SwcNode()
            test_path_length.add_data(child_match.data)
            tmp_node = child_match.parent
            # print("child match id = {}".format(
            #     child_match.data.get_id()))
            done = (tmp_node == ancestor_match)
            while not done:
                if tmp_node in path_length_map.keys():
                    prev_path_length = path_length_map[tmp_node]
                    test_path_length.add_length(prev_path_length)
                    done = True
                else:
                    test_path_length.add_length(tmp_node.data)
                    tmp_node = tmp_node.parent
                    if tmp_node is None:
                        done = True

                if tmp_node == ancestor_match:
                    done = True

            if tmp_node is None:
                if DEBUG:
                    print("[info]: descendant not match")
            else:
                path_length_map[child_match] = test_path_length
                if ancestor_trajectory.get_x() == -1.0 or ancestor_trajectory.get_z() == -1.0:
                    spe_ancestor_trajectory = diadam_match_utils.get_trajectory_for_path(ancestor_node, gold_node)
                    if ancestor_trajectory.get_x != -1.0:
                        spe_ancestor_trajectory.set_x(ancestor_trajectory.get_x())
                        spe_ancestor_trajectory.set_y(ancestor_trajectory.get_y())
                    if ancestor_trajectory.get_z != -1.0:
                        spe_ancestor_trajectory.set_z(ancestor_trajectory.get_z())

                test_xy_path_length = test_path_length.xy_path_length \
                                      + diadam_match_utils.get_end_node_XY_dis_diff(ancestor_node.data,
                                                                                    spe_ancestor_trajectory,
                                                                                    ancestor_match.data) \
                                      + diadam_match_utils.get_end_node_XY_dis_diff(gold_node.data,
                                                                                    des_tra,
                                                                                    child_match.data)
                test_z_path_length = test_path_length.z_path_length \
                                      + diadam_match_utils.get_end_node_Z_dis_diff(ancestor_node.data,
                                                                                   spe_ancestor_trajectory,
                                                                                   ancestor_match.data) \
                                      + diadam_match_utils.get_end_node_Z_dis_diff(gold_node.data,
                                                                                   des_tra,
                                                                                   child_match.data)

                if diadam_match_utils.path_length_matches(gold_path_length,
                                                          test_xy_path_length,
                                                          test_z_path_length) < 1:
                    return child_match

        if gold_node.has_children() and gold_node not in g_matches:
            stack.put(gold_node.left_son)
            stack.put(gold_node.right_son)

    return None


def is_sub_continuation(gold_node, ancestor_node, ancestor_match,
                        is_left, gold_path_length, bin_test_list, add_to_list):
    if ancestor_match is None:
        return False

    path_length_map = {}
    path_length_map[gold_node] = gold_path_length

    if is_left:
        ancestor_tra = ancestor_node.data.left_trajectory
    else:
        ancestor_tra = ancestor_node.data.right_trajectory

    left_child_match = get_des_in_for_continuation(
        gold_node.left_son, ancestor_node, ancestor_match, ancestor_tra, path_length_map, bin_test_list
    )
    right_child_match = get_des_in_for_continuation(
        gold_node.right_son, ancestor_node, ancestor_match, ancestor_tra, path_length_map, bin_test_list
    )

    if left_child_match is not None and right_child_match is not None:
        common_ancestor = diadam_match_utils.LCA(left_child_match, right_child_match, ancestor_match)
        if common_ancestor is not None and diadam_match_utils.is_within_dis_match_threshold(common_ancestor, gold_node):
            g_matches[gold_node] = common_ancestor
            g_matches[common_ancestor] = gold_node
            if add_to_list:
                g_distance_match.append(gold_node)
        elif add_to_list:
            g_continuation.append(gold_node)
        return True
    elif left_child_match is not None or right_child_match is not None:
        if add_to_list:
            g_continuation.append(gold_node)
        return True
    return False


def is_continuation(gold_node, bin_test_list, add_to_list=True, DEBUG=False):
    if not gold_node.has_children():
        return False

    ancestor_node_matches = []
    gold_path_length = swc_node.SwcNode()

    # find the first matched ancestor of the gold_node
    ancestor_match = None
    ancestor_gold = gold_node.parent
    gold_path_length.path_length = gold_node.data.path_length
    gold_path_length.xy_path_length = gold_node.data.xy_path_length
    gold_path_length.z_path_length = gold_node.data.z_path_length
    is_match = False
    ancestor_match = False
    is_left = False

    if DEBUG:
        print("gold_path_length: path_length {}, xy_path_length {}, z_path_length {}".format(
            gold_path_length.path_length,
            gold_path_length.xy_path_length,
            gold_path_length.z_path_length
        ))
    while ancestor_gold is not None and not ancestor_match:
        if ancestor_gold in g_matches.keys():
            ancestor_match = g_matches[ancestor_gold]
            if is_sub_continuation(gold_node, ancestor_gold, ancestor_match,is_left,
                                   gold_path_length, bin_test_list, add_to_list):
                return True
            ancestor_match = True
        ancestor_node_matches = diadam_match_utils.get_nearby_node_list(ancestor_gold, bin_test_list, check_previous_use=True, g_matches=g_matches)
        for a_n_node in ancestor_node_matches:
            if is_sub_continuation(gold_node, ancestor_gold, ancestor_match, is_left,
                                   gold_path_length, bin_test_list, add_to_list):
                return True

        gold_path_length.path_length += ancestor_gold.data.path_length
        gold_path_length.xy_path_length += ancestor_gold.data.xy_path_length
        gold_path_length.z_path_length += ancestor_gold.data. z_path_length
        is_left = ancestor_gold.is_left()
        ancestor_gold = ancestor_gold.parent

    # failed
    return False


def get_excess_weight(excess, node, excess_map):
    if g_weight_node == WEIGHT_SQRT_DEGREE:
        return math.sqrt(excess)
    if g_weight_node == WEIGHT_DEGREE_HARMONIC_MEAN:
        return 2.0/(1.0/excess_map[node.left_son] + 1.0/excess_map[node.right_son])
    if g_weight_node == WEIGHT_UNIFORM:
        return excess > 0
    if g_weight_node == WEIGHT_DEGREE:
        return excess


def remove_nearest_node(node, nearby_nodes, bin_gold_list):
    closest_dis = -1
    closest_match = None

    for nearby_node in nearby_nodes:
        dis = node.data.distance(nearby_node.data)
        if closest_dis == -1 or dis < closest_dis:
            closest_dis = dis
            closest_match = nearby_node

    bin_gold_list.remove(closest_match)


def weight_excess(bin_gold_subroot, bin_test_subroot, DEBUG=False):
    global g_excess_nodes

    if DEBUG:
        for node in g_matches.keys():
            print("matches = {}".format(node.data.get_id()))

    sum_weight = 0.0
    gold_bin_list = []
    setup_stack = queue.LifoQueue()
    use_stack = queue.LifoQueue()

    for son in bin_gold_subroot:
        gold_bin_list.extend(son.get_node_list())

    for son in bin_test_subroot:
        setup_stack.put(son)
        use_stack.put(son)
        while not setup_stack.empty():
            node = setup_stack.get()
            if node.has_children():
                setup_stack.put(node.left_son)
                setup_stack.put(node.right_son)
                use_stack.put(node.left_son)
                use_stack.put(node.right_son)

        direct_term_excess = {}

        while not use_stack.empty():
            node = use_stack.get()
            data = node.data
            # if data.get_id()
            if DEBUG:
                print("{} sum weight {}".format(
                    data.get_id(),
                    sum_weight
                ))

            if node.has_children():
                num_excess_node = direct_term_excess[node.left_son]
                num_excess_node += direct_term_excess[node.right_son]
                excess_weight = get_excess_weight(num_excess_node, node, direct_term_excess)

                if not node in g_matches.keys() and \
                        len(diadam_match_utils.get_nearby_node_list(node, gold_bin_list,
                                                                    check_previous_use=True, g_matches=g_matches)) == 0 and \
                        not is_continuation(node, gold_bin_list, add_to_list=False):
                    g_excess_nodes[node] = excess_weight
                    if not node in g_spur_set:
                        sum_weight += excess_weight
                else:
                    num_excess_node = 0
                    excess_weight = 0
            else:
                num_excess_node = 0
                excess_weight = 0

                if node not in g_spur_set and node not in g_matches.keys() and node.parent not in g_matches.keys():
                    nearby_nodes = diadam_match_utils.get_nearby_node_list(node, gold_bin_list,
                                                                           check_previous_use=True, g_matches = g_matches)
                    if len(nearby_nodes) > 0:
                        remove_nearest_node(node, nearby_nodes, gold_bin_list)
                    else:
                        num_excess_node = 1
                        excess_weight = 1
                        g_excess_nodes[node] = 1
                        sum_weight += excess_weight

            direct_term_excess[node] = num_excess_node
    return sum_weight


def score_trees(bin_gold_root, bin_test_root, bin_gold_subroots, bin_test_subroots, DEBUG=False):
    global g_weight_dict
    global g_weight_sum
    global g_score_sum
    global g_quantity_score_sum
    global g_final_score
    global g_direct_match_score
    global g_quantity_score

    stack = queue.LifoQueue()
    number_of_nodes = 0
    weight = 0

    bin_test_list = []
    bin_gold_list = []

    # Root is certainly matched, note that here is a address use
    bin_gold_root.data.parent_trajectory = bin_gold_root.data.get_center()
    bin_gold_root.data.left_trajectory = bin_gold_root.data.get_center()
    bin_gold_root.data.right_trajectory = bin_gold_root.data.get_center()
    if bin_test_root is not None:
        bin_test_root.data.parent_trajectory = bin_test_root.data.get_center()
        bin_test_root.data.left_trajectory = bin_test_root.data.get_center()
        bin_test_root.data.right_trajectory = bin_test_root.data.get_center()

        g_matches[bin_gold_root] = bin_test_root
        g_matches[bin_test_root] = bin_gold_root

        for son in bin_test_subroots:
            son.parent = bin_test_root
            bin_test_list.extend(son.get_node_list())

    for son in bin_gold_subroots:
        son.parent = bin_gold_root
        bin_gold_list.extend(son.get_node_list())

    # Increment quantity (non-continuations)
    g_quantity_score_sum += 1

    number_of_nodes -= len(g_spur_set)
    number_of_nodes += len(bin_gold_list)

    for gold_subroot in bin_gold_subroots:
        stack.put(gold_subroot)

        while not stack.empty():
            gold_node = stack.get()
            gold_data = gold_node.data
            if DEBUG:
                print(gold_node.data.get_id())
                if gold_node.data.get_id() in [4,5,6,7]:
                    print("line:616")

            if gold_node.has_children():
                stack.put(gold_node.left_son)
                stack.put(gold_node.right_son)

            if gold_node in g_spur_set:
                pass
            else:
                weight = g_weight_dict[gold_node]
                g_weight_sum += weight
                match = get_closest_match(gold_node=gold_node,
                                          bin_gold_list=bin_gold_list,
                                          bin_test_list=bin_test_list)

                # another match method (temporarily undo)
                # if match is None and gold_node.is_leaf() and g_terminal_threshold > 0:
                #     match = get_extended_termination_match(gold_node = gold_node,
                #                                            bin_gold_list = bin_gold_list,
                #                                            bin_test_list = bin_test_list)

                if match is not None:
                    g_matches[gold_node] = match
                    g_matches[match] = gold_node
                    g_score_sum += weight

                    # Increment quantity (non-continuations)
                    g_quantity_score_sum += 1
                else:
                    g_miss.add(gold_node)
        if DEBUG:
            gold_subroot.print_tree()
            print("--END--\n")
            if gold_subroot in g_matches.keys() is not None:
                g_matches[gold_subroot].print_tree()
    t_remove = []

    for miss_node in g_miss:
        if miss_node.left_son is not None or miss_node.right_son is not None:
            if is_continuation(miss_node, bin_test_list):
                weight = g_weight_dict[miss_node]
                g_score_sum += weight
                if DEBUG:
                    print("[Info: ] find continuation in miss nodes {}".format(miss_node.to_str()))
                t_remove.append(miss_node)

    for node in t_remove:
        g_miss.remove(node)

    if g_weight_sum > 0:
        try:
            g_direct_match_score = g_quantity_score_sum / number_of_nodes
        except ZeroDivisionError:
            g_direct_match_score = 0
        try:
            g_quantity_score = g_score_sum / g_weight_sum
        except ZeroDivisionError:
            g_quantity_score = 0

        if g_remove_spur > 0:
            remove_spurs(bin_test_root)

        if g_count_excess_nodes:
            g_weight_sum += weight_excess(bin_gold_subroots, bin_test_subroots)

        try:
            g_final_score = g_score_sum / g_weight_sum
        except ZeroDivisionError:
            g_final_score = 0


def config_init(config):
    """
    Read data in config to global
    Args:
        None
    Returns:
        None
    """
    global g_remove_spur
    global g_align_tree_by_root
    global g_count_excess_nodes
    global g_weight_node
    global g_list_miss
    global g_list_distant_matches
    global g_list_continuations
    global g_find_proper_root

    if "remove_spur" in config:
        g_remove_spur = config["remove_spur"]
    if "align_tree_by_root" in config.keys():
        g_align_tree_by_root = config["align_tree_by_root"]
    if "count_excess_nodes" in config.keys():
        g_count_excess_nodes = config["count_excess_nodes"]
    if "weight_node" in config.keys():
        g_weight_node = config["weight_node"]
    if "list_miss" in config.keys():
        g_list_miss = config["list_miss"]
    if "list_distant_matches" in config.keys():
        g_list_distant_matches = config["list_distant_matches"]
    if "list_continuations" in config.keys():
        g_list_continuations = config["list_continuations"]
    if "find_proper_root" in config.keys():
        g_find_proper_root = config["find_proper_root"]


def diadem_init():
    """
    init global parameters of diadem metric
    Args:
        None
    Returns:
        None
    """
    global g_weight_sum
    global g_score_sum
    global g_quantity_score_sum
    global g_quantity_score
    global g_direct_match_score
    global g_final_score
    global g_miss
    global g_spur_set
    global g_matches
    global g_weight_dict
    global g_excess_nodes
    global g_distance_match
    global g_continuation

    g_weight_sum = 0
    g_score_sum = 0
    g_quantity_score_sum = 0
    g_quantity_score = 0
    g_direct_match_score = 0
    g_final_score = -1

    g_miss = set()
    g_spur_set = set()
    g_matches = {}
    g_weight_dict = {}
    g_excess_nodes = {}
    g_distance_match = []
    g_continuation = []


def color_tree_only():
    if g_list_miss:
        if len(g_miss) > 0:
            for node in g_miss:
                # 9 means this node is missed
                node.data._type = 9
        if len(g_excess_nodes) > 0:
            for node in g_excess_nodes.keys():
                # 10 means this node is excessive
                node.data._type = 10

    if g_list_continuations:
        if len(g_continuation) > 0:
            for node in g_continuation:
                # 11 means this node is a continuation
                node.data._type = 11

    if g_list_distant_matches:
        if len(g_distance_match) > 0:
            for node in g_distance_match:
                # 12 means this node is a distant match
                node.data._type = 12


def print_result():
    start = time.time()
    print("g_weight_sum = {}".format(
        g_weight_sum
    ))
    print("g_score_sum = {}".format(
        g_score_sum
    ))
    print("g_final_score = {}".format(
        g_final_score
    ))
    print("time cost = {}".format(
        time.time() - start)
    )
    if g_list_miss:
        if len(g_miss) > 0:
            print("---Nodes that are missed (position and weight)---")
            for node in g_miss:
                print("node_ID = {} poi = {} weight = {}".format(
                    node.data.get_id(), node.data._pos, g_weight_dict[node]
                ))
                # 9 means this node is missed
                node.data._type = 9
            print("--END--")
        else:
            print("---Nodes that are missed:None---")

        print("")

        if len(g_excess_nodes) > 0:
            print("---extra Nodes in test reconstruction (position and weight)---")
            for node in g_excess_nodes.keys():
                print("node_ID = {} poi = {} weight = {}".format(
                    node.data.get_id(), node.data._pos, g_excess_nodes[node]
                ))
                # 10 means this node is excessive
                node.data._type = 10
        else:
            print("---extra Nodes in test reconstruction: None---")

    if g_list_continuations:
        print("")
        if len(g_continuation) > 0:
            print("---continuation Nodes (position and weight)---")
            for node in g_continuation:
                print("node_ID = {} poi = {} weight = {}".format(
                    node.data.get_id(), node.data._pos, g_weight_dict[node]
                ))
                # 11 means this node is a continuation
                node.data._type = 11
        else:
            print("---continuation Nodes None---")

    if g_list_distant_matches:
        print("")
        if len(g_distance_match) > 0:
            print("Distant Matches")
            for node in g_distance_match:
                print("node_ID = {} poi = {} weight = {}".format(
                    node.data.get_id(), node.data._pos, g_weight_dict[node]
                ))
                # 12 means this node is a distant match
                node.data._type = 12
        else:
            print("Distant Matches: none")


def adjust_root(swc_gold_tree, swc_test_tree,
                test_kdtree, test_pos_node_dict, t_matches):
    """
    find a common root for both gold and test trees
    Args:
        swc_gold_tree(Swc_Tree):
        swc_test_tree(Swc_Tree):
        test_kdtree(kd tree established based on test tree)
        test_pos_node_dict(Dict): a dict of test tree
            keys: tuple(x, y, z)
            value: Swc_Node in position of tuple
        t_matches(Dict):
            A match Dict of nodes in gold test
            key: Swc Node in gold tree
            value: Swc Node in test tree
    """
    swc_gold_list = swc_gold_tree.get_node_list()
    swc_test_list = swc_test_tree.get_node_list()

    gold_vis_list = np.zeros(shape=(len(swc_gold_list) + 10,))
    test_vis_list = np.zeros(shape=(len(swc_test_list) + 10,))

    for node in swc_gold_list:
        if node.is_virtual():
            continue
        nearby_nodes = swc_node.get_nearby_swc_node_list(gold_node=node, threshold=node.radius() / 2,
                                                         test_kdtree=test_kdtree, test_pos_node_dict=test_pos_node_dict)
        for t_node in nearby_nodes:
            if not gold_vis_list[node.get_id()] and not test_vis_list[t_node.get_id()]:
                t_matches[node] = t_node
                swc_gold_tree.change_root(node.get_id())
                swc_test_tree.change_root(t_node.get_id())

                for sub_node in PreOrderIter(node):
                    gold_vis_list[sub_node.get_id()] = 1
                for sub_node in PreOrderIter(t_node):
                    test_vis_list[sub_node.get_id()] = 1
                break


def diadem_metric(gold_swc_tree, test_swc_tree, config):
    """Main function of diadem metric
    Args:
        gold_swc_tree(SwcTree):
        test_swc_tree(SwcTree):
        config(Dict):
            The keys of 'config' is the name of configs, and the items are config values

    Example:
        test_tree = swc_node.SwcTree()
        gold_tree = swc_node.SwcTree()
        gold_tree.load("..\\..\\data\\test_data\\topo_metric_data\\sgold_fake_data2.swc")
        test_tree.load("..\\..\\data\\test_data\\topo_metric_data\\stest_fake_data2.swc")
        score, recall, precision = ssd_metric(gold_swc_tree=gold_tree,
                                              test_swc_tree=test_tree,
                                              config=config)

    Return:
        tuple: contain three values to demonstrate metric result
            g_weight_sum(float): weight of all nodes in the gold tree
            g_score_sum(float): weight of all the nodes that are matched.
            g_final_score(float): ratio of above two value

    Raise:
        None
    """
    global g_spur_set
    global g_weight_dict
    gold_swc_tree.set_node_type_by_topo(root_id=1)
    test_swc_tree.set_node_type_by_topo(root_id=5)
    diadem_init()
    config_init(config)
    diadam_match_utils.diadem_utils_init(config)
    scale = config["scale"]
    gold_swc_tree.rescale(scale)
    test_swc_tree.rescale(scale)
    test_kdtree, test_pos_node_dict = point_match_utils.create_kdtree(test_swc_tree.get_node_list())

    t_matches = {}
    debug = False
    if g_find_proper_root:
        adjust_root(swc_gold_tree=gold_swc_tree, swc_test_tree=test_swc_tree,
                    test_kdtree=test_kdtree, test_pos_node_dict=test_pos_node_dict,
                    t_matches=t_matches)
    if g_align_tree_by_root:
        test_swc_tree.align_roots(gold_tree=gold_swc_tree, t_matches=t_matches)

    # the id of root here is always -1, so we need to figure out first level sons of root
    for sub_gold_root in gold_swc_tree.root().children:
        if sub_gold_root in t_matches.keys():
            sub_test_root = t_matches[sub_gold_root]
            bin_test_root, bin_test_list = bin_utils.convert_to_binarytrees(sub_test_root)
        else:
            bin_test_root, bin_test_list = None, []

        bin_gold_root, bin_gold_list = bin_utils.convert_to_binarytrees(sub_gold_root)

        if g_remove_spur > 0:
            g_spur_set = remove_spurs(bin_gold_root, 1.0)

        generate_node_weights(bin_gold_root, bin_gold_list, g_spur_set)
        score_trees(bin_gold_root,
                    bin_test_root,
                    bin_gold_list,
                    bin_test_list)

        # debug
        if debug:
            for key in g_matches:
                print('match1 = {}, match2 = {}'.format(
                    key.data.get_id(), g_matches[key].data.get_id()
                ))
    color_tree_only()
    if debug:
        for k in g_weight_dict:
            print("id = {} wt = {}".format(k.data.get_id(), g_weight_dict[k]))

    if config['debug'] is True:
        print_result()

    res = {
        "weight_sum": g_weight_sum,
        "score_sum": g_score_sum,
        "final_score": g_final_score
    }
    return res, gold_swc_tree, test_swc_tree


def pyneval_diadem_metric(gold_swc, test_swc, config):
    """ interface to webmets, which is a web visualization project abandoned now
    Args: Same as func:diadem_metric
        gold_swc_tree(SwcTree):
        test_swc_tree(SwcTree):
        config(Dict):
            The keys of 'config' is the name of configs, and the items are config values
    Return:
        result(Dict): contain several metric results
    """
    gold_tree = swc_node.SwcTree()
    test_tree = swc_node.SwcTree()

    gold_tree.load_list(read_swc.adjust_swcfile(gold_swc))
    test_tree.load_list(read_swc.adjust_swcfile(test_swc))

    diadem_res = diadem_metric(gold_swc_tree=gold_tree,
                              test_swc_tree=test_tree,
                              config=config)

    result = {
        'gold_swc': gold_tree.to_str_list(),
        'test_swc': test_tree.to_str_list(),
        'weight_sum': diadem_res[0],
        'score_sum': diadem_res[1],
        'final_score': diadem_res[2]
    }

    return result


if __name__ == "__main__":
    start_time = time.time()
    testTree = swc_node.SwcTree()
    goldTree = swc_node.SwcTree()

    goldTree.load("../../data/test_data/topo_metric_data/gold_fake_data3.swc")
    testTree.load("../../data/test_data/topo_metric_data/test_fake_data3.swc")
    config_utils.get_default_threshold(goldTree)
    config = read_json.read_json("../../config/diadem_metric.json")
    config_schema = read_json.read_json("../../config/schemas/diadem_metric_schema.json")

    try:
        jsonschema.validate(config, config_schema)
    except Exception as e:
        raise Exception("[Error: ]Error in analyzing config json file")

    diadem_result, tmp1, tmp2 = diadem_metric(test_swc_tree=testTree,
                                              gold_swc_tree=goldTree,
                                              config=config)
    print("matched weight = {}\n"
          "total weight   = {}\n"
          "diadem score   = {}\n".
          format(diadem_result["weight_sum"], diadem_result["score_sum"], diadem_result["final_score"]))
